/**
 * \file
 *
 * \brief Low-level initialization functions called upon chip startup.
 *
 * Copyright(c) 2016 Atmel Corporation,
 *                    a wholly owned subsidiary of Microchip Technology Inc.
 *
 * \asf_license_start
 *
 * \page License
 *
 * Licensed under the Apache License, Version 2.0(the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the Licence at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * \asf_license_stop
 *
 */

#include "samd51.h"

/**
 * Initial system clock frequency.
 */
#define __SYSTEM_CLOCK   (120000000)

uint32_t SystemCoreClock = __SYSTEM_CLOCK;/*!< System Clock Frequency(Core Clock)*/

/**
 * Initialize the system
 *
 * @brief  Setup the microcontroller system.
 *         Initialize the System and update the SystemCoreClock variable.
 */
void SystemInit(void)
{
#if __FPU_USED
	/* Enable FPU */
	SCB->CPACR |= (0xFu << 20);
	__DSB();
	__ISB();
#endif

	// Enable XOSC32K clock
	OSC32KCTRL->XOSC32K.reg = OSC32KCTRL_XOSC32K_CGM_XT | OSC32KCTRL_XOSC32K_EN32K |
		OSC32KCTRL_XOSC32K_XTALEN | OSC32KCTRL_XOSC32K_ENABLE;
	while((OSC32KCTRL->STATUS.reg & OSC32KCTRL_STATUS_XOSC32KRDY) == 0);

	// Software reset
	GCLK->CTRLA.bit.SWRST = 1;
	while(GCLK->SYNCBUSY.reg & GCLK_SYNCBUSY_SWRST);

	// Connect OSCULP32K to clock generator 0
	GCLK->GENCTRL[0].reg = GCLK_GENCTRL_GENEN | GCLK_GENCTRL_SRC_OSCULP32K;
	while(GCLK->SYNCBUSY.reg & GCLK_SYNCBUSY_GENCTRL0);

	// Connect XOSC32K to clock generator 3
	GCLK->GENCTRL[3].reg = GCLK_GENCTRL_GENEN | GCLK_GENCTRL_SRC_XOSC32K;
	while(GCLK->SYNCBUSY.reg & GCLK_SYNCBUSY_GENCTRL3);

	// Configure DFLL48M
	OSCCTRL->DFLLCTRLA.reg = 0;
	OSCCTRL->DFLLMUL.reg = OSCCTRL_DFLLMUL_CSTEP(1) | OSCCTRL_DFLLMUL_FSTEP(1) |
		OSCCTRL_DFLLMUL_MUL(0);
	while(OSCCTRL->DFLLSYNC.reg & OSCCTRL_DFLLSYNC_DFLLMUL);

	OSCCTRL->DFLLCTRLB.reg = 0;
	while(OSCCTRL->DFLLSYNC.reg & OSCCTRL_DFLLSYNC_DFLLCTRLB);

	OSCCTRL->DFLLCTRLA.reg |= OSCCTRL_DFLLCTRLA_ENABLE;
	while(OSCCTRL->DFLLSYNC.reg & OSCCTRL_DFLLSYNC_ENABLE);

	OSCCTRL->DFLLVAL.reg = OSCCTRL->DFLLVAL.reg;
	while(OSCCTRL->DFLLSYNC.bit.DFLLVAL);

	OSCCTRL->DFLLCTRLB.reg = OSCCTRL_DFLLCTRLB_WAITLOCK | OSCCTRL_DFLLCTRLB_CCDIS |
		OSCCTRL_DFLLCTRLB_USBCRM ;
	while(!OSCCTRL->STATUS.bit.DFLLRDY);

	// Connect DFLL48M / 48 to clock generator 5 (1MHz)
	GCLK->GENCTRL[5].reg = GCLK_GENCTRL_DIV(48) | GCLK_GENCTRL_GENEN | GCLK_GENCTRL_SRC_DFLL;
	while(GCLK->SYNCBUSY.bit.GENCTRL5);

	// Setup DPLL0 to 120MHz
	GCLK->PCHCTRL[OSCCTRL_GCLK_ID_FDPLL0].reg = GCLK_PCHCTRL_CHEN | GCLK_PCHCTRL_GEN_GCLK5;
	OSCCTRL->Dpll[0].DPLLRATIO.reg = OSCCTRL_DPLLRATIO_LDRFRAC(0) | OSCCTRL_DPLLRATIO_LDR(119);
	while(OSCCTRL->Dpll[0].DPLLSYNCBUSY.bit.DPLLRATIO);

	OSCCTRL->Dpll[0].DPLLCTRLB.reg = OSCCTRL_DPLLCTRLB_LBYPASS | OSCCTRL_DPLLCTRLB_REFCLK_GCLK;
	OSCCTRL->Dpll[0].DPLLCTRLA.reg = OSCCTRL_DPLLCTRLA_ENABLE;
	while(OSCCTRL->Dpll[0].DPLLSTATUS.bit.CLKRDY == 0 || OSCCTRL->Dpll[0].DPLLSTATUS.bit.LOCK == 0);

	// Connect DPLL0 to clock generator 0 (120MHz)
	GCLK->GENCTRL[0].reg = GCLK_GENCTRL_DIV(1) | GCLK_GENCTRL_GENEN | GCLK_GENCTRL_SRC_DPLL0;
	while(GCLK->SYNCBUSY.reg & GCLK_SYNCBUSY_GENCTRL0);

	MCLK->CPUDIV.reg = MCLK_CPUDIV_DIV_DIV1;

	// Keep the default device state after reset
	SystemCoreClock = __SYSTEM_CLOCK;
	return;
}

/**
 * Update SystemCoreClock variable
 *
 * @brief  Updates the SystemCoreClock with current core Clock
 *         retrieved from cpu registers.
 */
void SystemCoreClockUpdate(void)
{
	// Not implemented
	SystemCoreClock = __SYSTEM_CLOCK;
	return;
}
