/**
 * @file spi_driver.c
 * @brief SPI driver
 *
 * @section License
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * Copyright (C) 2010-2019 Oryx Embedded SARL. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * @author Oryx Embedded SARL (www.oryx-embedded.com)
 * @version 1.9.6
 **/

//Dependencies
#include "same54.h"
#include "core/net.h"
#include "spi_driver.h"
#include "debug.h"

//SPI bitrate
#define SPI_BITRATE 5000000


/**
 * @brief SPI driver
 **/

const SpiDriver spiDriver =
{
   spiInit,
   spiSetMode,
   spiSetBitrate,
   spiAssertCs,
   spiDeassertCs,
   spiTransfer
};


/**
 * @brief SPI initialization
 * @return Error code
 **/

error_t spiInit(void)
{
   //Enable SERCOM0 core clock
   GCLK->PCHCTRL[SERCOM0_GCLK_ID_CORE].reg = GCLK_PCHCTRL_GEN_GCLK0 |
      GCLK_PCHCTRL_CHEN;

   //Enable PORT bus clock (CLK_PORT_APB)
   MCLK->APBBMASK.bit.PORT_ = 1;
   //Enable SERCOM0 bus clock (CLK_SERCOM_APB)
   MCLK->APBAMASK.bit.SERCOM0_ = 1;

   //Configure SCK pin (PB25)
   PORT->Group[1].PINCFG[25].bit.DRVSTR = 1;
   PORT->Group[1].PINCFG[25].bit.PMUXEN = 1;
   PORT->Group[1].PMUX[12].bit.PMUXO = MUX_PB25C_SERCOM0_PAD1;

   //Configure MOSI pin (PB24)
   PORT->Group[1].PINCFG[24].bit.DRVSTR = 1;
   PORT->Group[1].PINCFG[24].bit.PMUXEN = 1;
   PORT->Group[1].PMUX[12].bit.PMUXE = MUX_PB24C_SERCOM0_PAD0;

   //Configure MISO pin (PC25)
   PORT->Group[2].PINCFG[25].bit.DRVSTR = 1;
   PORT->Group[2].PINCFG[25].bit.PMUXEN = 1;
   PORT->Group[2].PMUX[12].bit.PMUXO = MUX_PC25C_SERCOM0_PAD3;

   //Configure CS pin (PA27)
   PORT->Group[0].DIRSET.reg = PORT_PA27;
   PORT->Group[0].OUTSET.reg = PORT_PA27;
   PORT->Group[0].PINCFG[27].bit.DRVSTR = 1;

   //Perform software reset
   SERCOM0->SPI.CTRLA.reg = SERCOM_SPI_CTRLA_SWRST;

   //Resetting the SERCOM0 (CTRLA.SWRST) requires synchronization
   while(SERCOM0->SPI.SYNCBUSY.reg & SERCOM_SPI_SYNCBUSY_SWRST);

   //Select master mode operation
   SERCOM0->SPI.CTRLA.reg = SERCOM_SPI_CTRLA_DIPO(3) |
      SERCOM_SPI_CTRLA_DOPO(0) | SERCOM_SPI_CTRLA_MODE(3);

   //Configure clock divider
   SERCOM0->SPI.BAUD.reg = SystemCoreClock / (2 * SPI_BITRATE) - 1;

   //Enable the receiver
   SERCOM0->SPI.CTRLB.reg = SERCOM_SPI_CTRLB_RXEN;

   //Writing to the CTRLB register when the SERCOM is enabled requires
   //synchronization
   while(SERCOM0->USART.SYNCBUSY.reg & SERCOM_USART_SYNCBUSY_CTRLB);

   //Enable SERCOM0
   SERCOM0->SPI.CTRLA.reg |= SERCOM_SPI_CTRLA_ENABLE;

   //Enabling and disabling the SERCOM requires synchronization
   while(SERCOM0->SPI.SYNCBUSY.reg & SERCOM_SPI_SYNCBUSY_ENABLE);

   //Successful processing
   return NO_ERROR;
}


/**
 * @brief Set SPI mode
 * @param mode SPI mode (0, 1, 2 or 3)
 **/

error_t spiSetMode(uint_t mode)
{
   //Not implemented
   return ERROR_NOT_IMPLEMENTED;
}


/**
 * @brief Set SPI bitrate
 * @param bitrate Bitrate value
 **/

error_t spiSetBitrate(uint_t bitrate)
{
   //Not implemented
   return ERROR_NOT_IMPLEMENTED;
}


/**
 * @brief Assert CS
 **/

void spiAssertCs(void)
{
   //Assert CS signal
   PORT->Group[0].OUTCLR.reg = PORT_PA27;
   //CS setup time
   usleep(1);
}


/**
 * @brief Deassert CS
 **/

void spiDeassertCs(void)
{
   //CS hold time
   usleep(1);
   //Deassert CS signal
   PORT->Group[0].OUTSET.reg = PORT_PA27;
   //CS disable time
   usleep(1);
}


/**
 * @brief Transfer a single byte
 * @param[in] data The data to be written
 * @return The data received from the slave device
 **/

uint8_t spiTransfer(uint8_t data)
{
   //Ensure the TX buffer is empty
   while(!SERCOM0->SPI.INTFLAG.bit.DRE);
   //Write character
   SERCOM0->SPI.DATA.bit.DATA = data;
   //Wait for the operation to complete
   while(!SERCOM0->SPI.INTFLAG.bit.RXC);

   //Return the received character
   return SERCOM0->SPI.DATA.bit.DATA;
}
